from typing import Any, IO, Text
import io

class GzipFile(io.BufferedIOBase):
    myfileobj: Any
    max_read_chunk: Any
    mode: Any
    extrabuf: Any
    extrasize: Any
    extrastart: Any
    name: Any
    min_readsize: Any
    compress: Any
    fileobj: Any
    offset: Any
    mtime: Any
    def __init__(self, filename: str = ..., mode: Text = ..., compresslevel: int = ...,
                 fileobj: IO[str] = ..., mtime: float = ...) -> None: ...
    @property
    def filename(self): ...
    size: Any
    crc: Any
    def write(self, data): ...
    def read(self, size=...): ...
    @property
    def closed(self): ...
    def close(self): ...
    def flush(self, zlib_mode=...): ...
    def fileno(self): ...
    def rewind(self): ...
    def readable(self): ...
    def writable(self): ...
    def seekable(self): ...
    def seek(self, offset, whence=...): ...
    def readline(self, size=...): ...

def open(filename: str, mode: Text = ..., compresslevel: int = ...) -> GzipFile: ...



from dataclasses import dataclass, asdict

import urwid


@dataclass(frozen=True)
class ComponentTheme:
    foreground: str = urwid.DEFAULT
    background: str = urwid.DEFAULT


@dataclass(frozen=True)
class Theme:
    body: ComponentTheme = ComponentTheme()
    foot: ComponentTheme = ComponentTheme(urwid.LIGHT_GRAY)
    focused: ComponentTheme = ComponentTheme(urwid.LIGHT_GRAY, urwid.DARK_BLUE)

    # auto-complete
    popup: ComponentTheme = ComponentTheme(urwid.BLACK, urwid.LIGHT_CYAN)
    popup_focused: ComponentTheme = ComponentTheme(
        urwid.WHITE, urwid.DARK_MAGENTA
    )

    # json
    json_key: ComponentTheme = ComponentTheme(urwid.LIGHT_BLUE)
    json_string: ComponentTheme = ComponentTheme(urwid.LIGHT_GREEN)
    json_integer: ComponentTheme = ComponentTheme(urwid.LIGHT_CYAN)
    json_numeric: ComponentTheme = ComponentTheme(urwid.LIGHT_CYAN)
    json_bool: ComponentTheme = ComponentTheme(urwid.YELLOW)
    json_null: ComponentTheme = ComponentTheme(urwid.LIGHT_RED)
    json_focused: ComponentTheme = ComponentTheme(
        urwid.LIGHT_GRAY, urwid.DARK_BLUE
    )

    def palette(self):
        palette = []
        for component_name, component_theme in asdict(self).items():
            component_name = component_name.replace('_', '.')
            palette.append([
                component_name,
                component_theme["foreground"],
                component_theme["background"]
            ])
        return palette



from typing import Any, Optional
from .pubkey import pubkey

class _DSAobj(pubkey):
    keydata: Any
    implementation: Any
    key: Any
    def __init__(self, implementation, key) -> None: ...
    def __getattr__(self, attrname): ...
    def sign(self, M, K): ...
    def verify(self, M, signature): ...
    def has_private(self): ...
    def size(self): ...
    def can_blind(self): ...
    def can_encrypt(self): ...
    def can_sign(self): ...
    def publickey(self): ...

class DSAImplementation:
    error: Any
    def __init__(self, **kwargs) -> None: ...
    def generate(self, bits, randfunc: Optional[Any] = ..., progress_func: Optional[Any] = ...): ...
    def construct(self, tup): ...

generate: Any
construct: Any
error: Any



from overrides import overrides

from ..json_widget import JSONWidget


class PrimitiveWidget(JSONWidget):
    """
    a widget to display JSON `string`, `integer`, `number`, `boolean`, `null` type
    """

    def __init__(self, node, display_key):
        super().__init__(node, False, display_key)

    @overrides
    def load_value_markup(self):
        return str(self.get_node().get_value())



import sys
def main():
    if len(sys.argv) > 1 and sys.argv[1] == "exec" and len(sys.argv) > 2:
        # Import everything from your package at the top level

        # Execute the command passed in the arguments
        import rp
        rp.exeval('from rp import *\n'+' '.join(sys.argv[2:]))
    else:
        # If no additional arguments or not the 'run' command, call _pterm
        from rp.r import _pterm
        _pterm()

if __name__ == '__main__':
    main()



from __future__ import unicode_literals

from six import text_type

from rp.prompt_toolkit.enums import IncrementalSearchDirection, SEARCH_BUFFER
from rp.prompt_toolkit.token import Token

from .utils import token_list_len
from .processors import Processor, Transformation

__all__ = (
    'DefaultPrompt',
)


class DefaultPrompt(Processor):
    """
    Default prompt. This one shows the 'arg' and reverse search like
    Bash/readline normally do.

    There are two ways to instantiate a ``DefaultPrompt``. For a prompt
    with a static message, do for instance::

        prompt = DefaultPrompt.from_message('prompt> ')

    For a dynamic prompt, generated from a token list function::

        def get_tokens(cli):
            return [(Token.A, 'text'), (Token.B, 'text2')]

        prompt = DefaultPrompt(get_tokens)
    """
    def __init__(self, get_tokens):
        assert callable(get_tokens)
        self.get_tokens = get_tokens

    @classmethod
    def from_message(cls, message='> '):
        """
        Create a default prompt with a static message text.
        """
        assert isinstance(message, text_type)

        def get_message_tokens(cli):
            return [(Token.Prompt, message)]
        return cls(get_message_tokens)

    def apply_transformation(self, cli, document, lineno, source_to_display, tokens):
        # Get text before cursor.
        if cli.is_searching:
            before = _get_isearch_tokens(cli)

        elif cli.input_processor.arg is not None:
            before = _get_arg_tokens(cli)

        else:
            before = self.get_tokens(cli)

        # Insert before buffer text.
        shift_position = token_list_len(before)

        # Only show the prompt before the first line. For the following lines,
        # only indent using spaces.
        if lineno != 0:
            before = [(Token.Prompt, ' ' * shift_position)]

        return Transformation(
                tokens=before + tokens,
                source_to_display=lambda i: i + shift_position,
                display_to_source=lambda i: i - shift_position)

    def has_focus(self, cli):
        # Obtain focus when the CLI is searching.

        # Usually, when using this `DefaultPrompt`, we don't have a
        # `BufferControl` instance that displays the content of the search
        # buffer. Instead the search text is displayed before the current text.
        # So, we can still show the cursor here, while it's actually not this
        # buffer that's focussed.
        return cli.is_searching


def _get_isearch_tokens(cli):
    def before():
        if cli.search_state.direction == IncrementalSearchDirection.BACKWARD:
            text = 'reverse-i-search'
        else:
            text = 'i-search'

        return [(Token.Prompt.Search, '(%s)`' % text)]

    def text():
        return [(Token.Prompt.Search.Text, cli.buffers[SEARCH_BUFFER].text)]

    def after():
        return [(Token.Prompt.Search, '`: ')]

    return before() + text() + after()


def _get_arg_tokens(cli):
    """
    Tokens for the arg-prompt.
    """
    arg = cli.input_processor.arg

    return [
        (Token.Prompt.Arg, '(arg: '),
        (Token.Prompt.Arg.Text, str(arg)),
        (Token.Prompt.Arg, ') '),
    ]



from dataclasses import dataclass

from overrides import overrides

from .query_bar import QueryBarKeys
from ...keymapper import AbstractComponentKeyMapper


@dataclass(frozen=True)
class QueryBarKeyMapper(AbstractComponentKeyMapper):

    query: str = "enter"
    cancel: str = "esc"

    @property
    @overrides
    def mapped_key(self):
        return {
            self.query: QueryBarKeys.QUERY,
            self.cancel: QueryBarKeys.CANCEL
        }



#Pychess, written by Ryan Burgert, 2018
#Run me in Python 3! There are no dependencies.
#For an extra boost in performance, try running me in pypy3!
init='''
♖ ♘ ♗ ♕ ♔ ♗ ♘ ♖
♙ ♙ ♙ ♙ ♙ ♙ ♙ ♙
· · · · · · · ·
· · · · · · · ·
· · · · · · · ·
· · · · · · · ·
♟ ♟ ♟ ♟ ♟ ♟ ♟ ♟
♜ ♞ ♝ ♛ ♚ ♝ ♞ ♜'''
#
vacant='·'
black='♙♖♘♗♕♔'
white='♟♜♞♝♛♚'
pawn  ='♟♙'
rook  ='♜♖'
knight='♞♘'
bishop='♝♗'
king  ='♚♔'
queen ='♛♕'
#
def group(*g):#g is vararg of strings
    from itertools import product#cartesian product
    return {x:y for x,y in product(''.join(g),g)if x in y}
piece   =group(pawn,rook,knight,bishop,king,queen)#{'♔':king,'♟':pawn,'♚':king, ... }
color   =group(black,white)#{'♔':black,'♘':black,'♚':white, ... }
value   ={pawn:1,rook:5,knight:3,bishop:3,king:1000,queen:9}
opponent={white:black,black:white}
#
def b2m(b):#board to matrix (make it mutable)
    b=b[1:]#get rid of the newline prefix
    return [x.split(' ')for x in b.split('\n')]
def m2b(m):#matrix to board (make it immutable)
    return '\n'+'\n'.join(' '.join(x)for x in m)
def move(b,x0,y0,x1,y1):#one-based from the bottom left
    assert legal(b, x0, y0, x1, y1),'illegal move'
    m=b2m(b)
    m[-y1][x1-1]=p=m[-y0][x0-1]
    assert p in piece,repr(p)+' is not a piece'
    m[-y0][x0-1]=vacant
    if p in pawn and y1 in{1,8}:#Not supported: promoting to non-queen
        m[-y1][x1-1],=set(queen)&set(color[p])#promote pawn to queen if pawn in a back row
    return m2b(m)
def flipb(b):#flip board:black/white's board positions
    return m2b(reversed([reversed(x)for x in b2m(b)]))
def flipm(*m):#flip move, *m in form x0,y0,x1,y1
    return (9-x for x in m)
def score(b,c):#the total value of all color c's pieces on board b
    return sum(value[piece[x]]for x in b if x in c)
def legal(b,x0,y0,x1,y1,print=lambda x:None):
    #Not supported: castling, en-passant
    m=b2m(b)
    if not 1<=x0<=8 or not 1<=y0<=8 or not 1<=x1<=8 or not 1<=y1<=8 or x0==x1 and y0==y1:#0 or 1 is off the board or piece didn't move
        print('Out of bounds or no movement')
        return False
    p0=m[-y0][x0-1]#piece before
    if p0 in black:
        m,x0,y0,x1,y1=b2m(flipb(b)),*flipm(x0,y0,x1,y1)#flip everything to make life easier
    try:
        p1=m[-y1][x1-1]#piece after
    except:
        print(-y1,x1-1)
        print(m)
        assert False,'Error?'
    Δ={abs(x1-x0),abs(y1-y0)}#In this function, the delta represents change in position
    if p0 in queen:#Queen can act as either bishop or rook
        print('Queen: Δ='+str(Δ))
        if 0 in Δ   :p0,=set(color[p0])&set(rook)
        elif len(Δ)==1:p0,=set(color[p0])&set(bishop)
        else:print('Queen neither acting as rook nor bishop, Δ='+str(Δ));return False
    if p0 in pawn:
        print('Pawn')
        if abs(x0-x1)>1:#Invalid horizontal movement
            print('Pawn cannot move horizontal')
            return False
        if x0==x1:#Not capturing a piece
            if y0==2 and y1==4:return p1 is vacant and m[-3][x0-1] is vacant#Move forward twice
            if y1==y0+1:       return p1 is vacant                          #Move forward once
            return False                                                    #Invalid move
        return y1==y0+1 and p1 in opponent[color[p0]]#Capturing a piece
    if p0 in rook:
        print('Rook')
        if x0==x1:#We moved on y axis
            x,={x0,x1}#basically, this says that x0 is equal to x1 and that x is just one of them. That way we don't have to choose arbitrarily, for more logically readable code.
            for y in range(min(y0,y1),max(y0,y1))[1:]:#make sure path BETWEEN 0 and 1 is vacant
                if m[-y][x-1] is not vacant:return False
            return p1 in vacant+opponent[color[p0]]#end must be opponent piece or vacant
        if y0==y1:#We moved on x axis
            y,={y0,y1}
            for x in range(min(x0,x1),max(x0,x1))[1:]:#make sure path BETWEEN 0 and 1 is vacant
                if m[-y][x-1] is not vacant:return False
            return p1 in vacant+opponent[color[p0]]#end must be opponent piece or vacant
        return False
    if p0 in knight:
        print('Knight')
        if not Δ=={1,2}:return False
        return p1 in vacant+opponent[color[p0]]
    if p0 in bishop:
        print('Bishop')
        if len(Δ)>1:return False#Not an equal difference
        Δ=Δ.pop()
        assert Δ,'Internal logic error: Piece should be guarenteed to move'
        for _ in range(Δ-1):
            x0+=1 if x0<x1 else -1
            y0+=1 if y0<y1 else -1
            if m[-y0][x0-1] is not vacant:return False#Must be clear path to target
        return p1 in vacant+opponent[color[p0]]
    if p0 in king:
        print('King')
        return Δ<={0,1} and p1 in vacant+opponent[color[p0]]
    print('Piece not recognized: '+repr(p0))
    return False
def moves(b,x,y):
    #Returns the set of all possible resulting legal boards from moving that piece
    p=b2m(b)[-y][x-1]
    Δ=set()#set of different movements to try
    r=set(range(-7,8))#max possible movement range on the board
    if p in pawn        :Δ|={(0,1),(0,2),(1,1),(-1,1),(0,-1),(0,-2),(1,-1),(-1,-1)}
    if p in rook  +queen:Δ|={(0,n)for n in r}|{( n,0)for n in r}
    if p in bishop+queen:Δ|={(n,n)for n in r}|{(-n,n)for n in r}
    if p in knight      :Δ|={(1,2),(1,-2),(-1,2),(-1,-2),( 2,1),( 2,-1),(-2, 1),(-2,-1)}
    if p in king        :Δ|={(1,0),(1, 1),( 0,1),(-1, 1),(-1,0),(-1,-1),( 0,-1),( 1,-1)}
    out=set()
    for Δx,Δy in Δ:
        args=b,x,y,x+Δx,y+Δy
        if legal(*args):
            yield move(*args)#it's safe to assume every element is unique
def shuffled(l):
    from random import shuffle
    l=list(l)+[]
    shuffle(l)
    return l
def all_moves(b,c):#board,color
    out=set()
    m=b2m(b)
    for x in shuffled(range(1,9)):
        for y in shuffled(range(1,9)):
            if m[-y][x-1] in c:
                for move in moves(b,x,y):
                    if move not in out:
                        yield move
                        out|={move}
def all_n_moves(b,c,n):
    #All combinations of n moves ahead of b, starting with color c
    out={b}
    for _ in range(n):
        new_out=out|set()
        for x in out:
            for move in all_moves(x,c):
                if move not in new_out:
                    yield move
                    new_out|={move}
        out=new_out
def advantage(b,c):
    return score(b,c)-score(b,opponent[c])
#
search=lambda b,c:all_n_moves(b,c,1)
def think_1(b,c,bad=9999):
    out,ba=b,-9999#biggest advantage seen
    for s in search(b,c):#for search-board in search; s is a string containing a chess board
        a=advantage(s,c)
        if a>ba:#Must not be a >=, must be >. If it's >=, then it searches for the best move you could make to help your opponent win.
            out=s
            ba=a
        if a>=bad:#The score is too high! Ignore it. Any further searching is a waste of resources. This is how we prune our search tree.
            return s,bad#This gets propagated through future calls such as in think_2 and think_3 etc,
    return out,ba#out is the actual chess board, while the second parameter is a number that measures how good the move is: in the form (move where c has most advantage). (c's advantage)
def think_2(b,c,bad=-9999):
    out,sa=b,9999#smallest advantage seen
    for s in search(b,c):#for search-board in search
        _,a=think_1(s,opponent[c],bad=sa)
        if a<sa:
            out=s
            sa=a
        if a<=bad:#Again, this is just here for performance improvements
            return s,bad
    return out,sa#move where c's opponent has least advantage; c's largest advantage = -(c's opponent's smallest advantage)
def think_3(b,c,bad=9999):
    out,ba=b,-9999
    for s in search(b,c):
        _,a=think_2(s,opponent[c],bad=ba)
        if a>ba:
            out=s
            ba=a
        if a>=bad:
            return s,bad
    return out,ba
def think_4(b,c,bad=-9999):
    out,sa=b,9999
    for s in search(b,c):
        _,a=think_3(s,opponent[c],bad=sa)
        if a<sa:
            out=s
            sa=a
        if a<=bad:
            return s,bad
    return out,sa
def has_won(b,c):
    return not set(b)&set(king)&set(opponent[c])#If our opponent color's king is nowhere to be found, we win!
#######################################################################################################################################################################################
#Nothing in the above code depends on the below code.
#######################################################################################################################################################################################

help_message="""PYCHESS:
    Welcome to Pychess, created by Ryan Burgert
    on planet Earth, Milky Way Galaxy, 2018.

    In this game, you play a chess game against an AI.
    There are a few differences in the rules between
    normal chess and the chess in this game, though.
        1. En-Passant and Castling are not allowed
        2. You are allowed to put yourself in check 
        3. There are no stalemates
        4. Pawns always promote to queens
        5. In order to win the game, you 
           must capture your opponent's king
        6. You always play white

    Special commands:
        1. "HELP" - brings up this menu
        2. "AUTO" - lets the AI make one move for you
        3. "L1"   - sets the AI to level 1
        4. "L2"   - sets the AI to level 2
        5. "L3"   - sets the AI to level 3
        6. "L4"   - sets the AI to level 4
        7. "L?"   - display current AI level
        8. "abcd", where a, b, c, and d are digits 
            between 1 and 8, inclusively - 
            Moves a piece from square (a,b) to square (c,d), 
            where coordinates (x,y) represent the positions
            of each square on the board. 
            EXAMPLE: To move your king's pawn two forward on 
            the first turn, you would enter "5254"
        9. The up/down arrow keys let you scroll
            through commands, for the sake of 
            convenience.
        10. "FLIP" - Flips the board visually.
            You still play white, though.
            This is useful for comparing 
            pychess to other chess engines.
        11. "SAVE *" - Saves the current game state to path *
            EXAMPLE: "SAVE test.txt"
        12. "LOAD *" - Loads the current game state from path *
            EXAMPLE: "LOAD test.txt"
        13. "EXIT" - Exits the game

    Other Notes: 
        • Commands are not case sensitive
        • Saved games include the ai level, and the sequence of
            moves since the beginning of the game. In other words,
            the UNDO command will work immediately after loading a game.
        • Immediately after you make a move, the chess board with
            that move (before your opponent makes their move)
            will be displayed in yellow
        • Your terminal must be capable of rendering unicode
            chess characters to display this game properly.
        • The AUTO command will make a move for you using the same
            AI level as your opponent AI, which will move immediately
            afterwards. This prevents cheating, as this function 
            is really meant to see how the AI thinks when playing 
            against itself."""
flip=False
def print_board(b,print=print):
    b=b.split('\n')[1:]
    h=' '
    for i in range(len(b)):
        b[i]=str(8-i)+h+b[i]
    b.append(h+' 1 2 3 4 5 6 7 8')
    if flip:
        b=[x[::-1]for x in b[::-1]]
    b='\n'+'\n'.join(b)
    print(b)
def color_print(s,n):
    print('\033[3'+str(n)+'m'+s+'\033[0m')#Using ansi-escape codes to change the output color in the terminal
def red_print(s):
    color_print(s,1)
def blue_print(s):
    color_print(s,6)#(Technically supposed to be cyan, but on my terminal it looks blue)
def yellow_print(s):
    color_print(s,3)
yellow_print(help_message)
game=[init]
game_over=lambda:has_won(game[-1],white) or has_won(game[-1],black)#Though usually "game over" means you lose by connotation, it doesn't technically mean that...
import readline#This lets us use the up/down arrow keys to navigate
if __name__=='__main__' or True:
    ai_level=4
    ai_move=lambda b,c=black:{1:think_1,2:think_2,3:think_3,4:think_4}[ai_level](b,c)[0]
    while True:
        if not game_over():
            print_board(game[-1])
        try:
            s=input(">>").strip()
            assert s,'Please input a command'
            if s.upper()=="UNDO":
                assert len(game)>1,'There are no moves left to UNDO, we are already at the beginning of the game'
                del game[-1]
                blue_print("Went back one move, it's your turn.")
            elif s.upper()=="HELP":
                yellow_print(help_message)
            elif s.upper()=="FLIP":
                blue_print("Flipping the board (for visual purposes only)...")
                flip^=True
            elif s.upper()=="EXIT":
                blue_print("Exiting the game...goodbye!")
                break
            elif s.upper().startswith("SAVE "):
                path=s[len("SAVE "):]
                file=open(path,'w')
                file.write(repr((ai_level,game)))
                file.close()
                blue_print("Saved the current game state to "+repr(path))
            elif s.upper().startswith("LOAD "):
                path=s[len("LOAD "):]
                file=open(path,'r')
                import ast
                try:
                    (ai_level,game)=ast.literal_eval(file.read())#It's like the eval function, except it's completely safe
                    blue_print("Loaded game state from "+repr(path))
                except:
                    blue_print("Error while loading game state! Failed to load "+repr(path))
                finally:
                    file.close()
            elif s.upper().startswith("L"):
                s=s[1:]
                assert s in {'1','2','3','4','?'},'Level must be either 1, 2, 3 or 4; or "?" to display current level'
                if s=='?':
                    blue_print("Current AI level: L"+str(ai_level))
                else:
                    ai_level=int(s)
                    blue_print("Set AI to level "+s)
            else:
                assert not game_over(),"The game's over - you can't make any more moves. You can still use other commands, though."
                if s.upper()=="AUTO":
                    blue_print("Letting the AI make this move for you...")
                    game.append(ai_move(game[-1],white))
                else:
                    assert len(s)==4 and all([x in '12345678' for x in s]),'Invalid move command! Your move command must be in the form "abcd", where a,b,c and d are digits between 1 and 8 inclusively.'
                    game.append(move(game[-1],*map(int,list(s))))
                print_board(game[-1],yellow_print)
                if has_won(game[-1],white):
                    blue_print("You win! You've saved humanity from this robot. Game over.")
                game[-1]=ai_move(game[-1])
                if has_won(game[-1],black):
                    blue_print("You Get Nothing! You Lose! Good Day, Sir! Game over.")
        except Exception as e:
            print(end='\007')#Ring the terminal's bell alert the user of an error
            red_print("ERROR: "+str(e))
assert False,'Game over. Re-import pychess to play again.'


"""
`Pyfx` TUI rendering related module.

It contains a homemade JSON widgets Library and collection of UI components to
interact with :class:`~pyfx.app.PyfxApp`.
"""
from .view_manager import View
from .view_config import ViewConfiguration



# Stubs for re
# Ron Murawski <ron@horizonchess.com>
# 'bytes' support added by Jukka Lehtosalo

# based on: http: //docs.python.org/2.7/library/re.html

from typing import (
    List, Iterator, overload, Callable, Tuple, Sequence, Dict,
    Generic, AnyStr, Match, Pattern, Any, Optional, Union
)

# ----- re variables and constants -----
DEBUG = 0
I = 0
IGNORECASE = 0
L = 0
LOCALE = 0
M = 0
MULTILINE = 0
S = 0
DOTALL = 0
X = 0
VERBOSE = 0
U = 0
UNICODE = 0
T = 0
TEMPLATE = 0

class error(Exception): ...

@overload
def compile(pattern: AnyStr, flags: int = ...) -> Pattern[AnyStr]: ...
@overload
def compile(pattern: Pattern[AnyStr], flags: int = ...) -> Pattern[AnyStr]: ...

@overload
def search(pattern: Union[str, unicode], string: AnyStr, flags: int = ...) -> Optional[Match[AnyStr]]: ...
@overload
def search(pattern: Union[Pattern[str], Pattern[unicode]], string: AnyStr, flags: int = ...) -> Optional[Match[AnyStr]]: ...

@overload
def match(pattern: Union[str, unicode], string: AnyStr, flags: int = ...) -> Optional[Match[AnyStr]]: ...
@overload
def match(pattern: Union[Pattern[str], Pattern[unicode]], string: AnyStr, flags: int = ...) -> Optional[Match[AnyStr]]: ...

@overload
def split(pattern: Union[str, unicode], string: AnyStr,
          maxsplit: int = ..., flags: int = ...) -> List[AnyStr]: ...
@overload
def split(pattern: Union[Pattern[str], Pattern[unicode]], string: AnyStr,
          maxsplit: int = ..., flags: int = ...) -> List[AnyStr]: ...

@overload
def findall(pattern: Union[str, unicode], string: AnyStr, flags: int = ...) -> List[Any]: ...
@overload
def findall(pattern: Union[Pattern[str], Pattern[unicode]], string: AnyStr, flags: int = ...) -> List[Any]: ...

# Return an iterator yielding match objects over all non-overlapping matches
# for the RE pattern in string. The string is scanned left-to-right, and
# matches are returned in the order found. Empty matches are included in the
# result unless they touch the beginning of another match.
@overload
def finditer(pattern: Union[str, unicode], string: AnyStr,
             flags: int = ...) -> Iterator[Match[AnyStr]]: ...
@overload
def finditer(pattern: Union[Pattern[str], Pattern[unicode]], string: AnyStr,
             flags: int = ...) -> Iterator[Match[AnyStr]]: ...

@overload
def sub(pattern: Union[str, unicode], repl: AnyStr, string: AnyStr, count: int = ...,
        flags: int = ...) -> AnyStr: ...
@overload
def sub(pattern: Union[str, unicode], repl: Callable[[Match[AnyStr]], AnyStr],
        string: AnyStr, count: int = ..., flags: int = ...) -> AnyStr: ...
@overload
def sub(pattern: Union[Pattern[str], Pattern[unicode]], repl: AnyStr, string: AnyStr, count: int = ...,
        flags: int = ...) -> AnyStr: ...
@overload
def sub(pattern: Union[Pattern[str], Pattern[unicode]], repl: Callable[[Match[AnyStr]], AnyStr],
        string: AnyStr, count: int = ..., flags: int = ...) -> AnyStr: ...

@overload
def subn(pattern: Union[str, unicode], repl: AnyStr, string: AnyStr, count: int = ...,
         flags: int = ...) -> Tuple[AnyStr, int]: ...
@overload
def subn(pattern: Union[str, unicode], repl: Callable[[Match[AnyStr]], AnyStr],
         string: AnyStr, count: int = ...,
         flags: int = ...) -> Tuple[AnyStr, int]: ...
@overload
def subn(pattern: Union[Pattern[str], Pattern[unicode]], repl: AnyStr, string: AnyStr, count: int = ...,
         flags: int = ...) -> Tuple[AnyStr, int]: ...
@overload
def subn(pattern: Union[Pattern[str], Pattern[unicode]], repl: Callable[[Match[AnyStr]], AnyStr],
         string: AnyStr, count: int = ...,
         flags: int = ...) -> Tuple[AnyStr, int]: ...

def escape(string: AnyStr) -> AnyStr: ...

def purge() -> None: ...
def template(pattern: Union[AnyStr, Pattern[AnyStr]], flags: int = ...) -> Pattern[AnyStr]: ...



from __future__ import unicode_literals, absolute_import
from rp.prompt_toolkit.contrib.regular_languages.compiler import compile

import re


def create_pdb_grammar(pdb_commands):
    """
    Create a compiled grammar for this list of PDB commands.

    (Note: this is an expensive function. Call only when `pdb_commands`
    changes.)
    """
    pdb_commands_re = '|'.join(map(re.escape, pdb_commands))

    def create_grammar(recursive=True):
        return r"""
        \s* (
            (?P<pdb_command>p|pp|whatis)     \s+  (?P<python_code>.*) |
            (?P<pdb_command>enable)          \s+  (?P<disabled_breakpoint>.*) |
            (?P<pdb_command>disable)         \s+  (?P<enabled_breakpoint>.*) |
            (?P<pdb_command>condition)       \s+  [0-9]+  \s+ (?P<python_code>.*)  |
            (?P<pdb_command>ignore)          \s+  (?P<breakpoint>[0-9]+)  \s+ [0-9]+  |
            (?P<pdb_command>commands)        \s+  (?P<breakpoint>[0-9]+)  |
            (?P<pdb_command>alias)           \s+  [^\s]+   \s+  """ + (create_grammar(False) if recursive else '.+') + """
            (?P<pdb_command>unalias)         \s+  (?P<alias_name>.*) |
            (?P<pdb_command>h|help)          \s+  (?P<pdb_command>.*) |
            (?P<pdb_command>display)         \s+  (?P<python_code>.*) |

            # For the break command, do autocompletion on file and function names.
            # After the comma, do completion on python code.
            (?P<pdb_command>break|b|tbreak)  \s+
                        (
                                # Break on a <function>
                                (?P<python_function>[^\s:]+)              |

                                # Break on a <file>:<lineno>
                                (?P<python_file>(?![0-9])[^\s:]+):[^\s]+  |

                                # Break on a <lineno>
                                [0-9]+
                        )
                        \s* (, \s* (?P<python_code>.*))? |

            # Known PDB commands (autocompletion & highlighting of all commands)
            (?P<pdb_command>""" + pdb_commands_re + """)  (\s+  .*)? |

            # Starting with exclamation mark -> always Python code.
            !(?P<python_code>.*) |

            # When the input is no valid PDB command, we consider it Python code.
            # (We use negative lookahead to be sure it doesn't start with a pdb
            # command or an exclamation mark.)
#            (?P<python_code>(?!(""" + pdb_commands_re + """)(\s+|$))(?!\!).*) |
            (?P<python_code_highlight_only>(?!(""" + pdb_commands_re + """)(\s+|$))(?!\!).*) |
        ) \s*
        """
    return compile(create_grammar())



from overrides import overrides

from ..json_widget import JSONWidget


class ArrayStartWidget(JSONWidget):
    """
    display widget for JSON `array` type node
    """

    def __init__(self, node, display_key):
        super().__init__(node, True, display_key)

    @overrides
    def load_value_markup(self):
        return "["
